import aiohttp
import aiofiles

from typing import Optional

def create_form_data(files: list, key_value_data: dict) -> aiohttp.FormData:
    """
    Create an instance of `aiohttp.FormData` encoding files (with the
    right filename and content type) and key-value data, provided as a dict.

    :param files: list, asset files represented as tuple like
        (filename, filepath, mimetype)
    :param key_value_data: dict, representing key-value pairs to encode as
        form-data fields
    :return: FormData, data to be sent as form-data
    """
    data = aiohttp.FormData()
    for filename, filepath, mimetype in files:
        data.add_field(
            filename,
            open(filepath, 'rb'),
            filename=filename,
            content_type=mimetype)
    for key, value in key_value_data.items():
        data.add_field(key, value)
    return data


async def download_file(url: str, file_path: str):
    """
    Asynchronously downloads a file from a URL.

    This function sends a GET request to the specified URL and writes the response to a file at the specified path.

    :param url: str, the URL of the file to download.
    :param file_path: str, the path where the downloaded file should be saved.
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            if resp.status == 200:
                f = await aiofiles.open(file_path, mode='wb')
                await f.write(await resp.read())
                await f.close()


async def send_request(url: str,
                       method: Optional[str] = "POST",
                       request_timeout: Optional[int] = 500,
                       **kwargs) -> dict:
    """
    Asynchronously sends a request to a URL.

    This function sends a GET or POST request to the specified URL and returns the response. If the response content
    type is not "application/json", it raises an exception.

    :param url: str, the URL to send the request to.
    :param method: str, optional, the HTTP method to use. Defaults to "POST".
    :param request_timeout: int, optional, the total timeout for the request in seconds. Defaults to 500.
    :param kwargs: additional keyword arguments to pass to the request.
    :return: dict, the JSON response from the server encoded as a dict.
    :raises Exception: If the response content type is not "application/json".
    """
    timeout = aiohttp.ClientTimeout(total=request_timeout)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        if method == "GET":
            res = await session.get(url, **kwargs)
        else:
            res = await session.post(url, **kwargs)
        if res.content_type == "application/json":
            content = await res.json()
            if res.status in [200, 403]:
                return content
            if "code" in content and "message" in content:
                raise Exception(f"{content['code']}: {content['message']}.")
            else:
                raise Exception(f"Unexpected response: {content}.")
        else:
            raise Exception(f"Unexpected response content type: {res.content_type}.")


import time

from typing import Optional
from dotenv import load_dotenv

from ..utils import get_env, read_files
from .utils import create_form_data, send_request


class AWClient:
    """
    Anything World API Client

                        .-. .-.    _                                   .-.     .-.
                       .' `.: :   :_;                                  : :     : :
     .--.  ,-.,-..-..-.`. .': `-. .-.,-.,-. .--.   .-..-..-. .--. .--. : :   .-' :
    ' .; ; : ,. :: :; : : : : .. :: :: ,. :' .; :  : `; `; :' .; :: ..': :_ ' .; :
    `.__,_;:_;:_;`._. ; :_; :_;:_;:_;:_;:_;`._. ;  `.__.__.'`.__.':_;  `.__;`.__.'
                  .-. :                     .-. :
                  `._.'                     `._.'

   Provides an interface to the Anything World API. It allows to send requests to the
   API to animate 3D models and retrieve them once they are done.
   """

    # Stage names that identify the end of a process for a given endpoint name
    _finished_stages = {
        "animate": {
            # In these stages, basic formats are already generated
            "default": [
                "format_conversion",
                "thumbnails_generation",
                "migrate_animation_finished"
            ],
            # In these stages, extra formats are already generated
            "extra_formats": [
                "formats_conversion_finished"
            ]
        }
    }

    def __init__(self, api_key: Optional[str] = None):
        """
        Initialize an instance of AWClient.

        :param api_key: str, API key to use for requests. If not provided, it will
            be read from the environment variable AW_API_KEY.
        """
        load_dotenv()
        self.api_key = api_key if api_key else get_env('AW_API_KEY')
        self.api_url = get_env('AW_API_URL')
        self.polling_url = get_env('AW_POLLING_URL')
        try:
            mode = get_env('AW_MODE')
            self.is_staging = mode == 'staging'
        except Exception:
            self.is_staging = False


    def find_by_name(self, model_name: str) -> dict:
        """
        Queries the Anything World API for a given model name.

        :param model_name: str, the name of the model to query.
        :return: dict, the JSON response from the API decoded as a dict.
        """
        return self._anything(model_name=model_name)


    def find(self, search_query: str) -> dict:
        """
        Queries the Anything World API for a given search query.

        :param search_query: str, the search query to query.
        :return: dict, the JSON response from the API decoded as a dict.
        """
        return self._anything(search_query=search_query)


    def animate(
            self,
            files_dir: str,
            model_name: str,
            model_type: str,
            is_symmetric: bool = True) -> dict:
        """
        Sends a request to animate a model.

        This function reads files from the specified directory, creates form data from the files and additional data,
        and sends a POST request to the API to animate the model.

        :param files_dir: str, the directory where the files to be animated are located.
        :param model_name: str, the name of the model to be animated.
        :param model_type: str, the type of the model to be animated.
        :param is_symmetric: bool, optional, a flag indicating whether the model is symmetric. Defaults to True.

        :return: dict, the JSON response from the API decoded as a dict.
        """
        data = {
            "api_key": self.api_key,
            "model_name": model_name,
            "model_type": model_type,
            "symmetry": "true" if is_symmetric else "false"
        }
        form_data = create_form_data(read_files(files_dir), {})
        params = {"staging": "true"} if self.is_staging else {}
        return send_request(
            url=f"{self.api_url}/animate",
            method="POST",
            data=data,
            files=form_data,
            params=params
        )


    def get_animated_model(
            self,
            model_id: str,
            extra_formats: bool = False,
            waiting_time: Optional[int] = 5,
            verbose: Optional[bool] = False) -> dict:
        """
        Retrieves an animated model by polling.

        This function sends a GET request to the API to retrieve the animated model. It polls the API until the model
        reaches the finished stage expected for the /animate endpoint, waiting a specified amount of time
        between each request.

        The request will look if the model is done animated and with the
        basic formats only generated (`.glb`, `.fbx`). If you want to wait until all
        formats are generated as well (`.gltf`, `.dae`), set the
        `extra_formats` param to `True`.   

        :param model_id: str, the ID of the model to retrieve.
        :param extra_formats: bool, optional, a flag indicating if the model
            should have extra formats in the response or not. Defaults to False.
        :param waiting_time: int, optional, the amount of time to wait between each request in seconds. Defaults to 5.
        :param verbose: bool, optional, a flag indicating whether to print detailed information about each request.
            Defaults to False.
        :return: dict, the JSON response from the API decoded as a dict.
        """    
        expected_formats = "extra_formats" if extra_formats else "default"
        return self._get_model_by_polling(
            model_id,
            expected_stages=self._finished_stages["animate"][expected_formats],
            waiting_time=waiting_time,
            verbose=verbose)


    def is_animation_done(self, model_id: str, extra_formats: bool=False) -> bool:
        """
        Checks if the animation of a model is done.

        This function sends a request to the API to check if the animation of the specified model is done.

        The request will look if the model is done animated and with the
        basic formats only generated (`.glb`, `.fbx`). If you want to wait until all
        formats are generated as well (`.gltf`, `.dae`), set the
        `extra_formats` param to `True`.   

        :param model_id: str, the ID of the model to check.
        :param extra_formats: bool, optional, a flag indicating if the model
            should have extra formats in the response or not. Defaults to False.
        :return: bool, True if the animation is done, False otherwise.
        """
        return self._is_model_done(model_id, "animate", extra_formats)


    def get_model(self, model_id: str) -> dict:
        """
        Retrieves a model.

        This function sends a GET request to the API to retrieve the specified model. If the response is a list with
        one item, it returns the item. Otherwise, it returns the whole response.

        :param model_id: str, the ID of the model to retrieve.
        :return: dict, the JSON response from the API decoded as a dict.
        """
        params = {
            'key': self.api_key,
            'id': model_id,
            'stage': 'done',
        }
        if self.is_staging:
            params["staging"] = "true"
        res = send_request(
            url=self.polling_url,
            method="GET",
            params=params)
        if isinstance(res, list) and len(res) == 1:
            return res[0]
        return res


    def _anything(
            self,
            model_name: Optional[str] = None,
            search_query: Optional[str] = None,
            fuzzy: Optional[bool] = True,
            ) -> dict:
        """
        Queries the Anything World API for a given model name or search query.

        :param model_name: str, optional, the name of the model to query.
        :param search_query: str, optional, the search query to query.
        :param fuzzy: bool, optional (default: true), whether to perform fuzzy/approximate matching of
            the query with the database entries' fields, using a custom implementation
            of the edit distance algorithm.
        :return: dict, the JSON response from the API decoded as a dict.
        """
        # Required params
        data = {
            "key": self.api_key
        }
        if self.is_staging:
            data["staging"] = "true"
        # Optional params
        if model_name:
            data["name"] = model_name
        if search_query:
            data["search"] = search_query
        if fuzzy:
            data["fuzzy"] = str(fuzzy).lower()

        return send_request(
            url=f"{self.api_url}/anything",
            method="GET",
            params=data
        )


    def _get_model_by_polling(
            self,
            model_id: str,
            expected_stages: list,
            waiting_time: Optional[int] = 5,
            warmup_time: Optional[int] = 0,
            verbose: Optional[bool] = False) -> dict:
        """
        Retrieves a model by polling until it reaches the expected stage.

        This function sends a GET request to the API to retrieve the specified model. It polls the API until the model
        reaches the expected stage, waiting a specified amount of time between each request.

        :param model_id: str, the ID of the model to retrieve.
        :param expected_stages: list, the possible stages that the model is expected to reach to
            be considered done.
        :param waiting_time: int, optional, the amount of time to wait between each request in seconds. Defaults to 5.
        :param warmup_time: int, optional, the amount of time to wait before sending the first request in seconds.
            This is useful specially for users with low connectivity, to avoid unnecessary requests, given that for
            some models the API takes a while to the model ready. Defaults to 0.  
        :param verbose: bool, optional, a flag indicating whether to print detailed information about each request.
            Defaults to False.
        :return: dict, the JSON response from the API decoded as a dict.
        """
        if warmup_time > 0:
            time.sleep(warmup_time)
        attempt_count = 0

        while True:
            attempt_count += 1
            status_prefix = f"Polling attempt #{attempt_count}."
            try:
                model_json = self.get_model(model_id)
            except Exception as e:
                if verbose:
                    print(f"{status_prefix} Error: {e}")
                return
            if "stage" in model_json:
                model_stage = model_json["stage"]
                if model_stage in expected_stages:
                    if verbose:
                        print(f"{status_prefix} Done.")
                    return model_json
            else:
                if verbose:
                    print(f"{status_prefix} Model is not ready yet...")
            time.sleep(waiting_time)


    def _is_model_done(
            self,
            model_id: str,
            endpoint: str,
            extra_formats: bool=False) -> bool:
        """
        Check if a model is done for a given endpoint.

        :param model_id: str, the ID of the model to check.
        :param endpoint: str, the endpoint to check.
        :param extra_formats: bool, optional, a flag indicating if the model
            should have extra formats in the response or not. Defaults to False.
        :return: bool, True if the model is done, False otherwise.
        """
        required_formats = "extra_formats" if extra_formats else "default"
        expected_stages = self._finished_stages[endpoint][required_formats]
        res = self.get_model(model_id)
        if "stage" in res:
            return res["stage"] in expected_stages
        return False




import os

from typing import Optional


def get_extension(filename: str) -> str:
    """
    Get the file extension from a given filename.

    :param filename: str, File name to get extension from
    :return: str, The file extension
    """
    return filename.split(".")[-1].lower()


def get_mimetype(file_path: str) -> str:
    """
    Get mimetype by file extension.
    
    Note: this is not a reliable way to get mimetype, but it's good enough for our
    purposes without adding extra requirements.
    If you need a more reliable way, please use python-magic.

    :param file_path: str, path to file
    :return: str, mimetype of file
    """
    extensions = {
        "mtl": "text/plain",
        "obj": "text/plain",
        "txt": "text/plain",
        "gltf": "text/plain",
        "fbx": "application/octet-stream",
        "bin": "application/octet-stream",
        "dae": "application/xml",
        "glb": "model/gltf-binary",
        "zip": "application/zip",
        "jpeg": "image/jpeg",
        "jpg": "image/jpg",
        "png": "image/png",
        "bmp": "image/bmp",
        "gif": "image/gif",
        "tif": "image/tiff",
        "tiff": "image/tiff",
        "tga": "image/x-tga",
        "targa": "image/x-tga"
    }
    try:
        return extensions.get(get_extension(file_path))
    except IndexError:
        raise Exception(f"Could not get extension for file {file_path}")


def get_env(key: str) -> Optional[str]:
    """
    Retrieves the value of an environment variable.

    This function retrieves the value of the specified environment variable. If the environment variable is not set,
    it raises an exception.

    :param key: str, the name of the environment variable.
    :return: str, optional, the value of the environment variable, or None if the environment variable is not set.
    :raises Exception: If the environment variable is not set.
    """
    value = os.getenv(key)
    if not value:
        raise Exception(f"Environment variable {key} is not set")
    return value


def read_files(files_dir: str) -> list:
    """
    Read files from a directory (or single file), returning a tuple with
    filename, full path to file and its mimetype (inferred by its
    extension-only for now).

    :param files_dir: str, path to asset directory (or single asset file)
    :return: list, files as tuples (file name, local path, mime type)
    """
    files_to_get = []
    if os.path.isdir(files_dir):
        files_to_get = os.listdir(files_dir)
    else:
        files_to_get = [files_dir]
    files = []
    for filename in files_to_get:
        if filename == files_dir:
            file_path = filename
        else:
            file_path = os.path.join(files_dir, filename)
        mimetype = get_mimetype(file_path)
        files.append((os.path.basename(file_path), file_path, mimetype)) 
    return files




import pytest
import asyncio

from context import async_api

@pytest.mark.asyncio
async def test_async_find_by_name():
    client = async_api.AWClient()
    response = await client.find_by_name(model_name="my cat")
    assert response, 'No response from /anything endpoint'


@pytest.mark.asyncio
async def test_async_find():
    client = async_api.AWClient()
    response = await client.find(search_query="caat")
    assert response, 'No response from /anything endpoint'


@pytest.mark.asyncio
async def test_async_404_on_find():
    client = async_api.AWClient()
    try:
        _ = await client.find(search_query="ararate")
    except Exception as e:
        assert e, 'No exception raised for 404 search query'


@pytest.mark.asyncio
async def _test_async_send_example_cat_to_animate():
    client = async_api.AWClient()
    animate_response = await client.animate('./examples/cat', 'some_cat', 'cat', is_symmetric=True)
    assert animate_response, 'No response from /animate endpoint'
    model_id = animate_response["model_id"]
    is_finished = await client.is_animation_done(model_id)
    assert not is_finished, 'Model is finished before polling'

    animated_response = await client.get_animated_model(model_id)
    assert animated_response, 'No response from polling endpoint'

    is_finished = await client.is_animation_done(model_id)
    assert is_finished


@pytest.mark.asyncio
async def _test_async_send_example_cat_for_extra_formats_to_animate():
    client = async_api.AWClient()
    animate_response = await client.animate('./examples/cat', 'some_cat', 'cat', is_symmetric=True)
    assert animate_response, 'No response from /animate endpoint'
    model_id = animate_response["model_id"]
    is_finished = await client.is_animation_done(model_id, extra_formats=True)
    assert not is_finished, 'Model is finished before polling'

    animated_response = await client.get_animated_model(model_id, extra_formats=True)
    assert animated_response, 'No response from polling endpoint'

    is_finished = await client.is_animation_done(model_id, extra_formats=True)
    assert is_finished


import requests

from typing import Optional


def create_form_data(files: list, key_value_data: dict) -> dict:
    """
    Create a dict encoding files (with the
    right filename and content type) and key-value data, provided as a dict.

    :param files: list, asset files represented as tuple like
        (filename, filepath, mimetype)
    :param key_value_data: dict, representing key-value pairs to encode as
        form-data fields
    :return: dict, data to be sent as form-data
    """
    data = []
    for filename, filepath, mimetype in files:
        data.append(('files', (filename, open(filepath, 'rb'), mimetype)))
    for key, value in key_value_data.items():
        data[key] = (None, value)
    return data


def download_file(url: str, file_path: str):
    """
    Downloads a file from a URL.

    This function sends a GET request to the specified URL and writes the response to a file at the specified path.

    :param url: str, the URL of the file to download.
    :param file_path: str, the path where the downloaded file should be saved.
    """
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(file_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192): 
                f.write(chunk)


def send_request(url: str,
                       method: Optional[str] = "POST",
                       request_timeout: Optional[int] = 500,
                       **kwargs) -> dict:
    """
    Sends a request to a URL.

    This function sends a GET or POST request to the specified URL and returns the response. If the response content
    type is not "application/json", it raises an exception.

    :param url: str, the URL to send the request to.
    :param method: str, optional, the HTTP method to use. Defaults to "POST".
    :param request_timeout: int, optional, the total timeout for the request in seconds. Defaults to 500.
    :param kwargs: additional keyword arguments to pass to the request.
    :return: dict, the JSON response from the server encoded as a dict.
    :raises Exception: If the response content type is not "application/json".
    """
    if method == "GET":
        res = requests.get(url, **kwargs)
    else:
        res = requests.post(url, **kwargs)
    content_type = res.headers["content-type"]
    if "application/json" in content_type:
        content = res.json()
        if res.status_code in [200, 403]:
            return content
        if "code" in content and "message" in content:
            raise Exception(f"{content['code']}: {content['message']}.")
        else:
            raise Exception(f"Unexpected response: {content}.")
    else:
        raise Exception(f"Unexpected response content type: {content_type}.")


import pytest

from context import sync_api

def test_sync_find_by_name():
    client = sync_api.AWClient()
    response = client.find_by_name(model_name="my cat")
    assert response, 'No response from /anything endpoint'


def test_sync_find():
    client = sync_api.AWClient()
    response = client.find(search_query="caat")
    assert response, 'No response from /anything endpoint'


def test_sync_404_on_find():
    client = sync_api.AWClient()
    try:
        _ = client.find(search_query="ararate")
    except Exception as e:
        assert e, 'No exception raised for 404 search query'


def test_sync_send_example_cat_to_animate():
    client = sync_api.AWClient()
    animate_response = client.animate('./examples/cat', 'some_cat', 'cat', is_symmetric=True)
    print("animate response", animate_response)
    assert animate_response, 'No response from /animate endpoint'
    model_id = animate_response["model_id"]
    is_finished = client.is_animation_done(model_id)
    assert not is_finished, 'Model is finished before polling'

    animated_response = client.get_animated_model(model_id, verbose=True)
    assert animated_response, 'No response from polling endpoint'

    is_finished = client.is_animation_done(model_id)
    assert is_finished


def test_sync_send_example_cat_for_extra_formats_to_animate():
    client = sync_api.AWClient()
    animate_response = client.animate('./examples/cat', 'some_cat', 'cat', is_symmetric=True)
    print("animate response", animate_response)
    assert animate_response, 'No response from /animate endpoint'
    model_id = animate_response["model_id"]
    is_finished = client.is_animation_done(model_id, extra_formats=True)
    assert not is_finished, 'Model is finished before polling'

    animated_response = client.get_animated_model(model_id, extra_formats=True, verbose=True)
    assert animated_response, 'No response from polling endpoint'

    is_finished = client.is_animation_done(model_id, extra_formats=True)
    assert is_finished
